/* Nessus
 * Copyright (C) 2002 Michel Arboi
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * $Id: nessus-identd.c,v 1.2 2002/08/14 11:49:35 mikhail Exp $
 */

#include <includes.h>

static int
takelock()
{
  char	lockname[MAXPATHLEN];
  int	fd;

#ifndef NESSUSD_STATEDIR
# warning "NESSUSD_STATEDIR is undefined"
# define NESSUSD_STATEDIR "/tmp"
#endif
  snprintf(lockname, MAXPATHLEN, "%s/identdlock", NESSUSD_STATEDIR);
  fd = open(lockname, O_WRONLY|O_CREAT|O_TRUNC, 0600);
  if (fd < 0)
    {
      perror(lockname);
      return -1;
    }
  if (lockf(fd, F_TLOCK, 1) < 0)
    {
      perror(lockname);
      (void) close(fd);
      return -1;
    }
  return fd;
}

static int
init_socket_tcp(int port)
{
  struct sockaddr_in	addr;
  int			sock;


  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      perror("socket");
      return -1;
    }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(sock, &addr, sizeof(addr)) < 0)
    {
      perror("bind");
      /* No use to close socket */
      return -1;
    }

  if (listen(sock, 10) < 0)
    {
      perror("listen");
      return -1;
    }

  return sock;
}

static int
init_socket_unix(const char* socketpath)
{
  struct sockaddr_un	addr;
  int			sock;


  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      perror("socket");
      return -1;
    }

  memset(&addr, 0, sizeof(addr));
  addr.sun_family = AF_UNIX;
  strcpy(addr.sun_path, socketpath);
  if (bind(sock, &addr, sizeof(addr)) < 0)
    {
      perror("bind");
      /* No use to close socket */
      return -1;
    }

  if (listen(sock, 10) < 0)
    {
      perror("listen");
      return -1;
    }

  return sock;
}

#define MAX_CL	64	/* clients */

#define NI_RD	1		/* Reading the request */
#define NI_RD_E	2		/* Finish reading - Processing... */
#define NI_WR	3		/* Writing an answer */
#define NI_W_I	4		/* Waiting for peer information (ID...) */

struct {
  int		fd;		/* socket */
  short		st;		/* state */
  short		typ;		/* Type: 0 for Unix, 1 for remote */
  size_t	bufsz;		/* buffer size (malloced) */
  unsigned char	*buf;		/* buffer */
  int		ptr;		/* current R/W pointer */
  time_t	tictac;		/* Last IO date */
  int		srvP, clnP;	/* RFC 1413 server & client ports */
}	cli[MAX_CL];

struct {
  int		port;		/* TCP port */
  int		fd;		/* Socket */
  char		*name;		/* login name */
}	ids[MAX_CL];

/*
 * Dixit RFC 1413 :
 * The server should close the connection down after a configurable
 * amount of time with no queries - a 60-180 second idle timeout is
 * recommended.  The client may close the connection down at any time;
 * however to allow for network delays the client should wait at least
 * 30 seconds (or longer) after a query before abandoning the query and
 * closing the connection.
 */

#define DEF_TIMEOUT	60

static void
free_sock(int i)
{
  int	j, s = cli[i].fd;

  if (! cli[i].typ)
    for (j = 0; j < MAX_CL; j ++)
      if (ids[j].fd == s)
	{
	  ids[j].port = 0;
	  ids[j] = 0;
	  free(ids[j].name);
	  ids[j].name = NULL;
	}
  close(s);
  cli[i].fd = -1;
}

/*
 * TCP protocol: Cf. RFC 1413
 *
 * Unix protocol: all lines ends with \r\n
 * "Set name" : 'N' port pid name
 *    Answer : 1 : OK - 0 : error (malloc?)
 * "Asked?" : 'A'
 *    Answer : 1+ : yes - 0 : no
 */

static void
main_loop(int ts, int us)
{
  fd_set	rd, wr;
  int		i, j, n, maxfd, s, st, sz, len, x1, x2;
  unsigned char	line[80], *p, c;
  struct timeval	to;
  time_t	tictac;


  /* Init */
  for (i = 0; i < MAX_CL; i ++)
    cli[i].fd = -1;

  /* Main loop */

  for (;;)
    {
      FD_ZERO(&rd); FD_ZERO(&wr);
      FD_SET(ts, &rd); FD_SET(us, &rd);
      maxfd = MAX(ts, us);

      for (i = 0; i < MAX_CL; i ++)
	if (cli[i].fd >= 0)
	  {
	    if (cli[i].bufsz < 64)
	      {
		p = realloc(cli[i].buf, 64);
		if (p == NULL)
		  {
		    perror("realloc");
		    continue;
		  }
		else
		  {
		    cli[i].buf = p;
		    cli[i].bufsz = 64;
		  }
	      }
	    s = cli[i].fd; st = cli[i].st;
	    if (st == NI_RD)
	      FD_SET(s, &rd);
	    else if (st == NI_WR)
	      FD_SET(s, &wr); 
	    else 
	      s = -1;
	    if (s > maxfd)
	      maxfd = s;
	  }

      tictac = time(NULL);
      to.tv_sec = DEF_TIMEOUT; to.tv_usec = 0;
      if (select(maxfd+1, &rd, &wr, NULL, &to) < 0)
	{
	  perror("select");
	  return;
	}

      /* Incoming connections */
      
      if (FD_ISSET(us, &rd))
	{
	  if ((s = accept(us, NULL, NULL)) < 0)
	    perror("accept");
	  else
	    {
	      for (j = 0; j < MAX_CL; j ++)
		if (cli[j].fd < 0)
		  break;
	      if (j < MAX_CL)
		{
		  cli[j].fd = s;
		  cli[j].st = NI_RD;
		  cli[j].ptr = 0;
		  cli[j].tictac = tictac;
		  cli[j].typ = 0;
		}
	      else
		close(s);	/* Dirty but simple ;-) */
	    }
	}

      /* Duplicated code :-( */
      if (FD_ISSET(ts, &rd))
	{
	  if ((s = accept(ts, NULL, NULL)) < 0)
	    perror("accept");
	  else
	    {
	      for (j = 0; j < MAX_CL; j ++)
		if (cli[j].fd < 0)
		  break;
	      if (j < MAX_CL)
		{
		  cli[j].fd = s;
		  cli[j].st = NI_RD;
		  cli[j].ptr = 0;
		  cli[j].tictac = tictac;
		  cli[j].typ = 1; /* (!) */
		}
	      else
		close(s);
	    }
	}
      
      /* Established connections */

      for (i = 0; i < MAX_CL; i ++)
	{
	  s = cli[i].fd; st = cli[i].st;
	  if (s < 0)
	    continue;
	  if (st == NI_RD && FD_ISSET(s, &rd))
	    {
	      n = recv(s, line, sizeof(line), 0);
	      if (n < 0)
		{
		  if (errno != EINTR)
		    {
		      perror("recv");
		      free_sock(i);
		    }
		}
	      /* What if n == 0? */
	      else
		{
		  cli[i].tictac = tictac;
		  sz = n + cli[i].ptr;
		  if (sz >= cli[i].bufsz)
		    {
		      sz = MAX(sz + 1, cli[i].bufsz + 64);
		      /* If realloc(NULL) does not work, fix this code or
			 fix your OS! */
		      if ((p = realloc(cli[i].buf, sz)) == NULL)
			{
			  perror("realloc");
			  /* 
			   * RFC 1413 does not allow us to return something
			   * like "ENOMEM"
			   */
			  close(s);
			  cli[i].fd = -1;
			}
		      else
			cli[i].buf = p;
		    }
		  if (cli[i].fd > 0)
		    {
		      memcpy(cli[i].buf + cli[i].ptr, line, n);
		      cli[i].ptr += n;
		      cli[i].buf[cli[i].ptr] = '\0';
		      if (cli[i].ptr > 2 && 
			  strcmp(cli[i].buf + (cli[i].ptr - 2), "\r\n") == 0)
			cli[i].st = NI_RD_E;
		    }
		}
	    }
	  if (st == NI_WR && FD_ISSET(s, &wr))
	    {
	      p = cli[i].buf + cli[i].ptr;
	      len = strlen(p);
	      n = send(s, p, len, MSG_DONTWAIT);
	      if (n <= 0)
		{
		  switch (errno)
		    {
#if defined EWOULDBLOCK
		    case EWOULDBLOCK:
#elif define EAGAIN
		    case EAGAIN:
#endif
		    case EINTR:
		      break;

		    default:
		      free_sock(i);
		      break;
		    }
		}
	      else
		{
		  cli[i].tictac = tictac;
		  cli[i].ptr += n;
		}
	      
	    }

	  /* Process received requests */
	  for (i = 0; i < MAX_CL; i ++)
	    if (cli[i].fd >= 0 && cli[i].st == NI_RD_E)
	      {
		if (cli[i].typ)
		  {
		    /* TCP socket - parse RFC 1413 request */
		    if (sscanf(cli[i].buf, "%d,%d", &x1, &x2) != 2 || 
			x1 < 1 || x1 > 65535 || x2 < 1 || x2 > 65535)
		      {
			strcpy(cli[i].buf, "0,0: ERROR: INVALID-PORT\r\n");
			cli[i].st = NI_WR;
			continue;
		      }
		    cli[i].srvP = x1;
		    cli[i].clnP = x2;
		    for (j = 0; j < MAX_CL; j ++)
		      if (ids[j].port == x1)
			{
			  len = strlen(ids[j].name) + 32;
			  if (len > cli[i].bufsz)
			    {
			      p = realloc(cli[i].buf, len);
			      if (p == NULL)
				{
				  perror("realloc");
				}
			      cli[i].buf = p;
			      cli[i].bufsz = len;
			    }
			  if (len > cli[i].bufsz) /* realloc failed */
			    {
			      sprintf(cli[i].buf,
				      "%d, %d : ERROR : UNKNOWN-ERROR\r\n", 
				      cli[i].srvP, cli[i].clnP);
			    }
			  else
			    {
			      sprintf(cli[i].buf,
				      "%d, %d : USERID : UNIX : %s\r\n",
				      cli[i].srvP, cli[i].clnP, ids[j].name);
			    }
			  cli[i].ptr = 0;
			  cli[i].st = NI_WR;
			  break;
			}
		  }
		else
		  {
		    /* Unix socket */
		    c = cli[i].buf[0];
		    switch(c)
		      {
		      case 'N':	/* Set Name */
			break;
		      case 'A':	/* Asked? */
			break;
		      default:
			fprintf(stderr, "Unknown request '%c' (0x%x)\n",
				isprint(c) ? c : '.', c);
			break;
		      }
		    TBD;
		  }
	      }
	}
    }
  /*NOTREACHED*/
}


int
main(int argc, char* argv[])
{
  int		fd, port, s1, s2;
  char		*socketpath;
  
  if (argc != 3)
    {
      fprintf(stderr, "This program should be run by nessusd\n\
Usage: nessus-identd port socketpath\n");
      return 1;
    }

  if ((fd = takelock()) < 0)
    {
      fprintf(stderr, "Could not take lock\n");
      return 1;
    }
  
  port = atoi(argv[1]);
  socketpath = argv[2];

  if ((s1 = init_socket_tcp(port)) < 0)
    return 1;			/* Will close the file & release the lock */
  if ((s2 = init_socket_unix(socketpath)) < 0)
    return 1;			/* Will close the file & socket */
  main_loop(s1, s2);
  return 0;
}
